class Report:
    def show_2d_array(self, title, reader_file_list, column_width):
        max_length_title = 30
        title = title[0:(max_length_title - 1)]
        length_title = len(title)
        print("=== ", title, " ", end="")
        print("=" * (max_length_title - length_title))
        print()
        #
        for row in reader_file_list:
            for cell_value in row:
                #if isinstance(cell_value, str):
                if isinstance(cell_value, str) or isinstance(cell_value, int):
                    #print('{val:{fill}{width}}'.format(val=cell_value, fill='', width=column_width), end="  ")
                    print('{val:{fill}{width}}'.format(val=cell_value, fill='', width=column_width, left_aligned=True), end="  ")

            print()
        print()
        print()


    def show_comparation_result(self, row_number, column_number, actual_data, expected_data, error_message):
        print("Row: ", row_number + 1, "  Column: ", column_number + 1, "  =>  Actual data: ", actual_data, "    Expected data: ", expected_data, "    Remark / Error message: ", error_message)



# Script for comparing csv-files, 2d-array with a csv-file or 2d-arrays and for comparing text-files, text variable with a text-file or text variables.
#
import csv
import os

from DataComparerLibrary.arraycomparer import ArrayComparer


class DataComparer:
    def compare_data_2d_array_with_file(self, actual_data, expected_file, delimiter_expected_data=",", quotechar_expected_data='"', template_literals_dict=None):
        self.__check_if_actual_data_is_present(actual_data)
        self.__check_if_expected_file_is_present(expected_file)        
        #
        expected_data = self.__open_csv_input_file(expected_file, delimiter_expected_data, quotechar_expected_data)
        #
        ArrayComparer.compare_data(self, actual_data, expected_data, template_literals_dict)


    def compare_data_file_with_2d_array(self, actual_file, expected_data, delimiter_actual_data=",", quotechar_actual_data='"', template_literals_dict=None):
        self.__check_if_actual_file_is_present(actual_file)      
        self.__check_if_expected_data_is_present(expected_data)
        #
        actual_data = self.__open_csv_input_file(actual_file, delimiter_actual_data, quotechar_actual_data)
        #
        ArrayComparer.compare_data(self, actual_data, expected_data, template_literals_dict)


    def compare_data_2d_arrays(self, actual_data, expected_data, template_literals_dict=None):
        self.__check_if_actual_data_is_present(actual_data)
        self.__check_if_expected_data_is_present(expected_data) 
        #
        ArrayComparer.compare_data(self, actual_data, expected_data, template_literals_dict)


    def compare_data_files(self, actual_file, expected_file, delimiter_actual_data=",", delimiter_expected_data=",", quotechar_actual_data='"', quotechar_expected_data='"', template_literals_dict=None):
        self.__check_if_actual_file_is_present(actual_file) 
        self.__check_if_expected_file_is_present(expected_file)            
        #
        actual_data = self.__open_csv_input_file(actual_file, delimiter_actual_data, quotechar_actual_data)
        expected_data = self.__open_csv_input_file(expected_file, delimiter_expected_data, quotechar_expected_data)
        #
        ArrayComparer.compare_data(self, actual_data, expected_data, template_literals_dict)


    def compare_text_variable_with_text_file(self, actual_text, expected_file, template_literals_dict=None):
        self.__check_if_actual_data_is_present(actual_text)
        self.__check_if_expected_file_is_present(expected_file)        
        #
        actual_data = self.__split_text_into_textline_array(actual_text)
        expected_data = self.__split_textfile_into_textline_array(expected_file)
        #
        ArrayComparer.compare_data(self, actual_data, expected_data, template_literals_dict)


    def compare_text_file_with_text_variable(self, actual_file, expected_text, template_literals_dict=None):
        self.__check_if_actual_file_is_present(actual_file)      
        self.__check_if_expected_data_is_present(expected_text)
        #
        actual_data = self.__split_textfile_into_textline_array(actual_file)
        expected_data = self.__split_text_into_textline_array(expected_text)
        #
        ArrayComparer.compare_data(self, actual_data, expected_data, template_literals_dict)


    def compare_text_variables(self, actual_text, expected_text, template_literals_dict=None):
        self.__check_if_actual_data_is_present(actual_text)
        self.__check_if_expected_data_is_present(expected_text) 
        #
        actual_data = self.__split_text_into_textline_array(actual_text)
        expected_data = self.__split_text_into_textline_array(expected_text)        
        #
        ArrayComparer.compare_data(self, actual_data, expected_data, template_literals_dict)


    def compare_text_files(self, actual_file, expected_file, template_literals_dict=None):
        self.__check_if_actual_file_is_present(actual_file) 
        self.__check_if_expected_file_is_present(expected_file)            
        #
        actual_data = self.__split_textfile_into_textline_array(actual_file)        
        expected_data = self.__split_textfile_into_textline_array(expected_file)
        #
        ArrayComparer.compare_data(self, actual_data, expected_data, template_literals_dict)



    def __check_if_actual_data_is_present(self, data):
        if data == None:
            raise Exception("Actual Input data unknown.")        


    def __check_if_expected_data_is_present(self, data):
        if data == None:
            raise Exception("Expected Input data unknown.")        


    def __check_if_actual_file_is_present(self, file):
        if os.path.exists(file):
            print("actual_file: ", file)
        else:
            raise Exception("Actual Input file doesn't exists: ", file)


    def __check_if_expected_file_is_present(self, file):
        if os.path.exists(file):
            print("expected_file: ", file)    
        else:
            raise Exception("Expected Input file doesn't exists: ", file)  
        

    def __open_csv_input_file(self, input_file, delimiter_data=",", quotechar_data='"'):
        with open(input_file, mode='rt', encoding='utf-8') as input_file:
            if len(delimiter_data) == 1:
                data = list(csv.reader(input_file, delimiter=delimiter_data, quotechar=quotechar_data))
            else:
                data = list(csv.reader((line.replace(delimiter_data, chr(255)) for line in input_file), delimiter=chr(255), quotechar=quotechar_data))
        #
        return data        
    

    def __split_text_into_textline_array(self, text):
        data = []
        for line in text.split('\n'):
            data.append(line.strip('\n').split(chr(255)))
        #
        return data
    

    def __split_textfile_into_textline_array(self, input_file):
        with open(input_file, mode='rt', encoding='utf-8') as input_file:
            data = []
            for line in input_file.readlines():
                data.append(line.strip('\n').split(chr(255)))
        #
        return data
    


import datetime
import fnmatch
import re
import dateutil.relativedelta

from DataComparerLibrary.report import Report

class ArrayComparer:
    def compare_data(self, actual_data, expected_data_including_templates, template_literals_dict):
        difference_found = False
        #
        if actual_data and type(actual_data[0]) is not list:  # only a single row
            actual_data = [actual_data,]  # add row to tuple of lenght 1
        #
        if expected_data_including_templates and type(expected_data_including_templates[0]) is not list:  # only a single row
            expected_data_including_templates = [expected_data_including_templates,]  # add row to tuple of lenght 1

        number_of_rows_actual_data = len(actual_data)
        number_of_rows_expected_data = len(expected_data_including_templates)

        number_of_rows = number_of_rows_actual_data if (number_of_rows_actual_data >= number_of_rows_expected_data) else number_of_rows_expected_data

        Report.show_2d_array(self, "Actual data", actual_data, 20)
        Report.show_2d_array(self, "Expected data", expected_data_including_templates, 20)

        print()
        print("=== Overview differences between actual and expected data")
        print()

        for row_nr in range(number_of_rows):
            if row_nr >= number_of_rows_actual_data:
                difference_found = True
                if len(expected_data_including_templates[row_nr]) == 0:
                    Report.show_comparation_result(self, row_nr, 0, "", "", "Row actual data is not PRESENT. Row expected data is EMPTY.")
                else:
                    Report.show_comparation_result(self, row_nr, 0, "", expected_data_including_templates[row_nr][0], "Row actual data is not PRESENT.")
                continue
            #
            if row_nr >= number_of_rows_expected_data:
                difference_found = True
                if len(actual_data[row_nr]) == 0:
                    Report.show_comparation_result(self, row_nr, 0, "", "", "Row actual data is EMPTY. Row expected data is not PRESENT.")
                else:
                    Report.show_comparation_result(self, row_nr, 0, actual_data[row_nr][0], "", "Row expected data is not PRESENT.")
                continue
            #
            number_of_columns_actual_data   = len(actual_data[row_nr])
            number_of_columns_expected_data = len(expected_data_including_templates[row_nr])

            number_of_columns = number_of_columns_actual_data if (number_of_columns_actual_data >= number_of_columns_expected_data) else number_of_columns_expected_data

            for column_nr in range(number_of_columns):
                expected_data_including_date_template = None
                expected_data_with_wildcard = None
                skip_exception_rule_used = False
                #
                if column_nr >= number_of_columns_actual_data:
                    difference_found = True
                    Report.show_comparation_result(self, row_nr, column_nr, "", expected_data_including_templates[row_nr][column_nr], "Column actual data is not PRESENT.")
                    continue
                #
                if column_nr >= number_of_columns_expected_data:
                    difference_found = True
                    Report.show_comparation_result(self, row_nr, column_nr, actual_data[row_nr][column_nr], "", "Column expected data is not PRESENT.")
                    continue
                #
                if actual_data[row_nr][column_nr] != expected_data_including_templates[row_nr][column_nr]:
                    # Replace literal templates with fixed external strings.
                    if template_literals_dict:
                        for i in range(0, len(template_literals_dict)):
#                           key = list(template_literals_dict.keys())[i]
#                           value = list(template_literals_dict.values())[i]
#                           print("key: ", key)
#                           print("value: ", value)
                            expected_data_including_templates[row_nr][column_nr] = expected_data_including_templates[row_nr][column_nr].replace(list(template_literals_dict.keys())[i], list(template_literals_dict.values())[i])
#                           print("actual_data[row_nr][column_nr]: \n", actual_data[row_nr][column_nr])
#                           print("expected_data_including_templates[row_nr][column_nr]: \n", expected_data_including_templates[row_nr][column_nr])


                    # Verify if difference is a matter of string versus integer representation.
                    if str(actual_data[row_nr][column_nr]) == str(expected_data_including_templates[row_nr][column_nr]):
                        if isinstance(actual_data[row_nr][column_nr], int) and isinstance(expected_data_including_templates[row_nr][column_nr], str):
                            difference_found = True
                            Report.show_comparation_result(self, row_nr, column_nr, actual_data[row_nr][column_nr], expected_data_including_templates[row_nr][column_nr], "There is a difference between actual and expected data. Actual data is an integer while expected data is a string.")
                        elif isinstance(actual_data[row_nr][column_nr], str) and isinstance(expected_data_including_templates[row_nr][column_nr], int):
                            difference_found = True
                            Report.show_comparation_result(self, row_nr, column_nr, actual_data[row_nr][column_nr], expected_data_including_templates[row_nr][column_nr], "There is a difference between actual and expected data. Actual data is a string while expected data is an integer.")
                        continue
                    #
                    # If data in actual and expected field doesn't match, check if a template has been used in expected data.
                    match expected_data_including_templates[row_nr][column_nr].upper():
                        case "{PRESENT}":
                            if not actual_data[row_nr][column_nr]:
                                # No data is present in actual data field.
                                difference_found = True
                                Report.show_comparation_result(self, row_nr, column_nr, actual_data[row_nr][column_nr], expected_data_including_templates[row_nr][column_nr], "Actual data field is not PRESENT")
                        #
                        case "{EMPTY}":
                            if actual_data[row_nr][column_nr]:
                                # Actual data field is not empty.
                                difference_found = True
                                Report.show_comparation_result(self, row_nr, column_nr, actual_data[row_nr][column_nr], expected_data_including_templates[row_nr][column_nr], "Actual data field is not EMPTY")
                        #
                        case "{INTEGER}":
                            if isinstance(actual_data[row_nr][column_nr], int):
                                # A real integer.
                                continue
                            #
                            # Verify if string is integer.
                            if not actual_data[row_nr][column_nr].isdigit():
                                # Not positive integer field.
                                difference_found = True
                                Report.show_comparation_result(self, row_nr, column_nr, actual_data[row_nr][column_nr], expected_data_including_templates[row_nr][column_nr], "Actual data field is not INTEGER.")
                        #
                        case "{SKIP}":
                            pass
                        case _:
                            if "{SKIP}" in expected_data_including_templates[row_nr][column_nr].upper() or "{DATETIME_FORMAT():YYYYMMDDHHMMSSFF6}" in expected_data_including_templates[row_nr][column_nr].upper():
                                if expected_data_including_templates[row_nr][column_nr].upper() == "{SKIP}":
                                    # Complete actual data field will be skipped for verification.
                                    pass
                                else:
                                    # Part(s) of the actual data field will be skipped for verification.
                                    # Replace {SKIP}, ignoring cases, by wildcard *.
                                    # compiled = re.compile(re.escape("{SKIP}"), re.IGNORECASE)
                                    # expected_data_with_wildcard = compiled.sub("*", expected_data_including_templates[row_nr][column_nr])
                                    compiled = re.compile(re.escape("{SKIP}"), re.IGNORECASE)
                                    compiled2 = re.compile(re.escape("{DATETIME_FORMAT():YYYYMMDDHHMMSSFF6}"), re.IGNORECASE)
                                    expected_data_with_wildcard = compiled2.sub("*", compiled.sub("*", expected_data_including_templates[row_nr][column_nr]))
                                    #
                                    if fnmatch.fnmatch(actual_data[row_nr][column_nr], expected_data_with_wildcard):
                                        skip_exception_rule_used = True
                                        continue
                            #
                            if expected_data_with_wildcard is None:
                                # Wildcards not used.
                                expected_data_including_date_template = expected_data_including_templates[row_nr][column_nr]
                            else:
                                expected_data_including_date_template = expected_data_with_wildcard
                            #
                            if "{NOW()" in expected_data_including_templates[row_nr][column_nr].upper():
                                matches = ["{NOW():", "{NOW()+", "{NOW()-"]
                                if all([x not in expected_data_including_templates[row_nr][column_nr].upper() for x in matches]):
                                    difference_found = True
                                    Report.show_comparation_result(self, row_nr, column_nr, actual_data[row_nr][column_nr], expected_data_including_templates[row_nr][column_nr], "NOW() has been found in expected data field, but format is incorrect.")
                                    continue
                                #
                                expected_data = ArrayComparer.__replace_date_template_in_expected_data(self, expected_data_including_date_template)
                                #
                                if expected_data == -1:
                                    difference_found = True
                                    Report.show_comparation_result(self, row_nr, column_nr, actual_data[row_nr][column_nr], expected_data_including_templates[row_nr][column_nr], "NOW() has been found in expected data field, but format is incorrect.")
                                else:
                                    if not fnmatch.fnmatch(actual_data[row_nr][column_nr], expected_data):
                                        # No match despite using of wildcard(s).
                                        difference_found = True
                                        Report.show_comparation_result(self, row_nr, column_nr, actual_data[row_nr][column_nr], expected_data_including_templates[row_nr][column_nr], "Date template format displayed. See also next message line.")
                                        Report.show_comparation_result(self, row_nr, column_nr, actual_data[row_nr][column_nr], expected_data, "There is a difference between actual and expected data.")
                                continue
                                #
                            elif "{NOT(" in expected_data_including_templates[row_nr][column_nr].upper():
                                try:
                                    unwanted_expected_data = ArrayComparer.__get_unwanted_expected_data(self, expected_data_including_date_template)
                                    #
                                    if actual_data[row_nr][column_nr] == unwanted_expected_data:
                                        # Unwanted match.
                                        difference_found = True
                                        Report.show_comparation_result(self, row_nr, column_nr, actual_data[row_nr][column_nr], expected_data_including_templates[row_nr][column_nr], "NOT() template format displayed. See also next message line.")
                                        Report.show_comparation_result(self, row_nr, column_nr, actual_data[row_nr][column_nr], unwanted_expected_data, "Actual and expected data are equal. However actual data should NOT be equal to the expected data!!!")
                                except Exception as exception_message:
                                    # print(f"An exception occurred: {exception_message}")
                                    difference_found = True
                                    Report.show_comparation_result(self, row_nr, column_nr, actual_data[row_nr][column_nr], expected_data_including_templates[row_nr][column_nr], "NOT() has been found in expected data field, but format is incorrect.")
                                #
                            else:
                                if not skip_exception_rule_used:
                                    # No exceptions.
                                    difference_found = True
                                    Report.show_comparation_result(self, row_nr, column_nr, actual_data[row_nr][column_nr], expected_data_including_templates[row_nr][column_nr], "There is a difference between actual and expected data. No exception rule has been used.")
                            #
        if difference_found:
            print("\n\n\n")
            raise Exception("There is a difference between actual and expected data. See detail information.")
        else:
            print("There are no differences between actual and expected data found.")
            print("\n\n\n")


    def __replace_date_template_in_expected_data(self, expected_data_field_including_date_template):
        # Replace date_template in expected data.
        # For example: This is text {NOW()-5Y2M1D:YYYY-MM-DD} and also text.  =>  This is text 2018-05-03 and also text.
        position_open_brace_today_text  = expected_data_field_including_date_template.find("{NOW()")
        position_close_brace_today_text = expected_data_field_including_date_template.find("}", position_open_brace_today_text)
        #
        if position_close_brace_today_text == -1:
            return -1
        # Close brace of TODAY has been found.
        #
        expected_datetime_template_string = expected_data_field_including_date_template[position_open_brace_today_text:position_close_brace_today_text + 1]
        expected_datetime_string = ArrayComparer.__convert_datetime_template_to_datetime(self, expected_datetime_template_string)
        #
        if expected_datetime_string == -1:
            return -1
        # Datetime_template_string has been converted to datetime.
        #
        # Replace expected_datetime_template_string by expected_datetime_string in expected_data_field_including_template.
        compiled = re.compile(re.escape(expected_datetime_template_string), re.IGNORECASE)
        expected_data_with_calculated_date = compiled.sub(expected_datetime_string, expected_data_field_including_date_template)
        #
        return expected_data_with_calculated_date


    def __convert_datetime_template_to_datetime(self, expected_datetime_format):
        # Convert expected datetime template into datetime.
        # For example: {NOW():YYYY-MM-DD}         =>  2023-07-04
        #              {NOW():MMDDYY}             =>  070423
        #              {NOW()-5Y3M1D:D-MMMM-YY}   =>  3-April-18
        #              {NOW()-5Y2M1D:YYYY-MMM-DD} =>  2018-Apr-03
        #              {NOW()-5Y2M1D:YYYYMMDD}    =>  20180503
        #              {NOW()-5Y2M1D:YYYY-M-D}    =>  2018-5-3
        #              {NOW()+2D:DDMMYYYY         =>  06072023
        #              {NOW()-5Y2M1D:YYYY-MM-DD}  =>  2018-05-03
        template_datetime_string_splitted = expected_datetime_format.split(":")
        #
        match len(template_datetime_string_splitted):
            case 2:
                if template_datetime_string_splitted[0] == "{NOW()":
                    # Current date time.
                    expected_datetime = datetime.datetime.now()
                else:
                    # Adjust date time based on current date time.
                    relative_datetime_template_string = template_datetime_string_splitted[0].replace('{NOW()', '')
                    relative_datetime = ArrayComparer.__convert_relative_datetime_template_to_relative_datetime(self, relative_datetime_template_string[1:len(relative_datetime_template_string)])
                    if relative_datetime == -1:
                        return -1
                    else:
                        match relative_datetime_template_string[0]:
                            case "+":
                                expected_datetime = datetime.datetime.now() + relative_datetime
                            case "-":
                                expected_datetime = datetime.datetime.now() - relative_datetime
                            case _:
                                return -1
            case _:
                return -1
        #
        year = expected_datetime.strftime("%Y")
        year_2_digits = expected_datetime.strftime("%y")
        month = expected_datetime.strftime("%m")
        month_abbreviated = expected_datetime.strftime("%b")
        month_full = expected_datetime.strftime("%B")
        day = expected_datetime.strftime("%d")
        #
        expected_date = template_datetime_string_splitted[1][0:len(template_datetime_string_splitted[1]) - 1].replace("YYYY", year).replace("YY", year_2_digits).replace("MMMM", month_full).replace("MMM", month_abbreviated).replace("MM", month).replace("M", month.lstrip("0")).replace("DD", day).replace("D", day.lstrip("0"))
        return expected_date


    def __convert_relative_datetime_template_to_relative_datetime(self, relative_datetime_str):
        # Convert relative datetime template to relative datetime.
        # For example: 2Y5M1D  =>  2 years, 5 months, 1 day   (used to add to or subtract from current moment / date)
        # \d+ means 1 of more digits; search on character - for example Y;
        regex = re.compile(r'((?P<years>\d+?)Y)?((?P<months>\d+?)M)?((?P<days>\d+?)D)?((?P<hours>\d+?)h)?((?P<minutes>\d+?)m)?((?P<seconds>\d+?)s)?')
        period = regex.match(relative_datetime_str)

        if not period:
            return -1

        period = period.groupdict()
        kwargs = {}
        period_names = ["years", "months", "days"]
        #
        for name, param in period.items():
            if param:
                period_name = name
                period_count = param
                #
                if period_name in period_names:
                    kwargs[period_name] = int(period_count)
        #
        if kwargs:
            return dateutil.relativedelta.relativedelta(**kwargs)
        else:
            return -1



    def __get_unwanted_expected_data(self, expected_data_field_including_date_template):
        position_open_brace = expected_data_field_including_date_template.find("{NOT(")
        position_close_brace = expected_data_field_including_date_template.find(")}", position_open_brace)
        #
        if position_open_brace == -1:
            #print("position_open_brace:", position_open_brace)
            raise Exception()
        #
        if position_close_brace == -1:
            #print("position_close_brace:", position_close_brace)
            raise Exception()
        #
        unwanted_expected_data = expected_data_field_including_date_template[position_open_brace+5:position_close_brace]
        #
        if ArrayComparer.is_integer(self, unwanted_expected_data):
            unwanted_expected_data = int(unwanted_expected_data)
        return unwanted_expected_data



    def is_integer(self, string):
        if string[0] == '-':
            # if a negative number
            return string[1:].isdigit()
        else:
            return string.isdigit()





# Script for conversion on an input file. The converted result will be written to an output file.
# The files are open in binary mode.
# Currently, methods are implemented to remove or replace a separate linefeed within a record. Records are ended by carriage return linefeed.
#
import os
import re


class FileConverter:
    def remove_separate_lf(self, input_file, output_file):
        # Remove separate linefeed (LF), so carriage return linefeed (CRLF) remains.
        try:
            if not os.path.exists(input_file):
                raise Exception("Input file doesn't exists: ", input_file)
            #
            print("input_file: ", input_file)
            print("output_file: ", output_file)
            #
            with open(input_file, "rb") as input_file:
                with open(output_file, "wb") as output_file:
                    output_file.write(re.sub(b"(?<!\r)\n", b"", input_file.read()))
            #
        except Exception as error:
            raise Exception("Error message: ", type(error).__name__, "–", error)


    def replace_separate_lf(self, input_file, output_file, replacement_string, encoding_replacement_string='utf-8'):
        # Replace separate linefeed (LF), so carriage return linefeed (CRLF) remains.
        try:
            if not os.path.exists(input_file):
                raise Exception("Input file doesn't exists: ", input_file)
            #
            print("input_file: ", input_file)
            print("output_file: ", output_file)
            print("replacement_string: ", replacement_string)
            print("encoding: ", encoding_replacement_string)
            #
            with open(input_file, "rb") as input_file:
                with open(output_file, "wb") as output_file:
                    output_file.write(re.sub(b"(?<!\r)\n", replacement_string.encode(encoding_replacement_string), input_file.read()))
            #
        except Exception as error:
            raise Exception("Error message: ", type(error).__name__, "–", error)


    def remove_cr_and_lf(self, input):
        # Remove carriage return (CR) and linefeed (LF) from input string.
        print("input: ", input)
        output = str(input).replace("\r", "").replace("\n", "")
        print("output: ", output)
        #
        return output


    def replace_cr_and_lf(self, input, replacement_string):
        # Replace carriage return (CR) and linefeed (LF) from input string.
        print("input: ", input)
        output = str(input).replace("\r", replacement_string).replace("\n", replacement_string)
        print("output: ", output)
        #
        return output



from DataComparerLibrary import DataComparer

# Input files
actual_file   = "C:\\Users\\USER\\Documents\\Tool projecten\\Robot Framework Verzamelmap\\RobotDemo-master-Probeersel\\data_vergelijker\\actueel_resultaat\\actual.csv"
expected_file = "C:\\Users\\USER\\Documents\\Tool projecten\\Robot Framework Verzamelmap\\RobotDemo-master-Probeersel\\data_vergelijker\\verwacht_resultaat\\expected_equal.csv"


a = DataComparer()
a.compare_data_files(actual_file, expected_file, delimiter_actual_data=';', delimiter_expected_data=';')
a.compare_data_files(actual_file, expected_file, delimiter_actual_data=';', delimiter_expected_data=';', quotechar_actual_data=None, quotechar_expected_data=None)
a.compare_data_files(actual_file, expected_file)
actual_data_input = [['3', '1.1', '4', '7', '2023-05-27 12:00'],
                     ['1', '0', '2', '9', '2023-05-29 12:00'],
                     ['', '2', '0', '1', '5'],
                     ['5.1', '8', '1', '0', '3'],
                     ['3', '', '2023-06-19', '3', '2023-06-08']]

a.compare_data_files(actual_data_input, expected_file)

a.compare_data_files(actual_file, expected_file)




class Tools:
    def is_integer(self, string):
        if string[0] == '-':
            # if a negative number
            return string[1:].isdigit()
        else:
            return string.isdigit()



#!/usr/bin/env python

from .datacomparer import DataComparer
from .datasorter import DataSorter
from .fileconverter import FileConverter


class DataComparerLibrary(DataComparer, DataSorter, FileConverter):
    ROBOT_LIBRARY_SCOPE = 'GLOBAL'






class DelimiterTranslator(object):
    def __init__(self, file, old, new):
        self.file = file
        self.old = old
        self.new = new

    def write(self, s):
        self.file.write(s.replace(self.old, self.new))

    def close(self):
        self.file.close()

    def flush(self):
        self.file.flush()



#!/usr/bin/env python

"""Setup script for DataComparerLibrary distributions"""

import os
import re
import sys
from distutils.core import setup
from os.path import join, dirname, abspath
from pathlib import Path
from setuptools import setup

sys.path.insert(0, os.path.join('src', 'DataComparerLibrary'))

# read the contents of your README file
this_directory = Path(__file__).parent
README = (this_directory / "README.rst").read_text()

with open(join(this_directory, 'src', 'DataComparerLibrary', 'version.py')) as f:
    VERSION = re.search("\nVERSION = '(.*)'", f.read()).group(1)

def main():
    setup(name='DataComparerLibrary',
          version=VERSION,
          description="For comparing csv-files, 2d-array with a csv-file or 2d-arrays. For comparing text-files, text variable with a text-file or text variables. Including a sorting module.",
          long_description=README,
          long_description_content_type="text/x-rst",
          url="",
          author="René Philip Zuijderduijn",
          author_email="datacomparerlibrary@outlook.com",
          license="Apache",
          classifiers=[
              "License :: OSI Approved :: Apache Software License",
              "Operating System :: Microsoft :: Windows :: Windows 10",
              "Operating System :: Microsoft :: Windows :: Windows 11",
              "Programming Language :: Python :: 3.7",
              "Programming Language :: Python :: 3.8",
              "Programming Language :: Python :: 3.9",
              "Programming Language :: Python :: 3.10",
              "Programming Language :: Python :: 3.11",
              "Programming Language :: Python :: 3.12",
              "Programming Language :: Python :: 3.13",                            
          ],
          keywords='robotframework testing test-automation datacompare',
          package_dir={'': 'src'},
          packages=['DataComparerLibrary'],
          )


if __name__ == "__main__":
    main()



# Script for sorting an input file. Sorted result will be written to an output file. Encoding is UFT-08.
#
import csv
import os
from DataComparerLibrary.delimitertranslator import DelimiterTranslator


class DataSorter:
    def sort_csv_file(self, input_file, output_file, number_of_header_lines=0, number_of_trailer_lines=0, sort_on_columns_list=None, delimiter=',', quotechar=None):
        try:
            if not os.path.isfile(input_file):
                raise Exception("Input file doesn't exists: ", input_file)
            #
            print("input_file: ", input_file)
            print("output_file: ", output_file)
            print("number_of_header_lines: ", number_of_header_lines)
            print("number_of_trailer_lines: ", number_of_trailer_lines)
            print("sort_on_columns_list: ", sort_on_columns_list)
            print("delimiter: ", delimiter)
            print("quotechar: ", quotechar)
            #
            number_of_records_input_file = 0
            with open(input_file) as input_file_row_count:
                number_of_records_input_file = sum(1 for line in input_file_row_count)
            #
            print("number_of_records_input_file: ", number_of_records_input_file)
            #
            number_of_header_lines = int(number_of_header_lines)
            number_of_trailer_lines = int(number_of_trailer_lines)
            #
            if number_of_records_input_file < number_of_header_lines + number_of_trailer_lines:
                raise Exception("The number of records in the input file is less than the declared number of header and trailer lines.")
            #
            number_of_sort_columns = 0
            if sort_on_columns_list:
                number_of_sort_columns = len(sort_on_columns_list)
            print("number_of_sort_columns: ", number_of_sort_columns)
            #
            with open(input_file, mode='rt', encoding='utf-8') as input_file, open(output_file, mode='w', newline='', encoding='utf-8') as output_file:
                if len(delimiter) == 1:
                    input_data = list(csv.reader(input_file, delimiter=delimiter, quotechar=quotechar))
                    output_data = csv.writer(output_file, delimiter=delimiter, quotechar=quotechar)
                else:
                    input_data = list(csv.reader((line.replace(delimiter, chr(255)) for line in input_file), delimiter=chr(255), quotechar=quotechar))
                    translator = DelimiterTranslator(output_file, chr(255), delimiter)
                    output_data = csv.writer(translator, delimiter=chr(255), quotechar=quotechar)
                #
                data_text = []
                trailer_text = []
                #
                if number_of_header_lines == 0 and number_of_trailer_lines == 0:
                    # Only data lines.
                    data_text = input_data
                else:
                    # Header or/and trailer or/and data lines present.
                    line_nr = 0
                    #
                    # Separate in header, data and trailer text.
                    for row in input_data:
                        line_nr += 1
                        if number_of_header_lines != 0 and line_nr <= number_of_header_lines:
                            # A header line.
                            output_data.writerow(row)
                        elif number_of_trailer_lines != 0 and line_nr > number_of_records_input_file - number_of_trailer_lines:
                            # A trailer line.
                            trailer_text.append(row)
                        else:
                            # A data line.
                            data_text.append(row)
                #
                sorted_data_text = DataSorter.__sort_data_records(self, data_text, sort_on_columns_list)
                #
                for row in sorted_data_text:
                    output_data.writerow(row)

                for row in trailer_text:
                    output_data.writerow(row)
            #
        except IndexError as error:
            raise Exception("Probably selected column doesn't exist. Correct argument 'sort_on_columns_list'. Error message: ", type(error).__name__, "–", error)
        except Exception as error:
            raise Exception("Error message: ", type(error).__name__, "–", error)



    def __sort_data_records(self, data_text, sort_on_columns_list):
        number_of_sort_columns = 0
        if sort_on_columns_list:
            number_of_sort_columns = len(sort_on_columns_list)
        print("number_of_sort_columns: ", number_of_sort_columns)
        #
        match number_of_sort_columns:
            case 0:
                sorted_data_text = sorted(data_text, key=lambda column: (column[0]))
            case 1:
                a = sort_on_columns_list[0]
                sorted_data_text = sorted(data_text, key=lambda column: (int(column[a]) if isinstance(a, int) else column[int(a)]))
            case 2:
                a = sort_on_columns_list[0]
                b = sort_on_columns_list[1]
                sorted_data_text = sorted(data_text, key=lambda column: (int(column[a]) if isinstance(a, int) else column[int(a)], int(column[b]) if isinstance(b, int) else column[int(b)]))
            case 3:
                a = sort_on_columns_list[0]
                b = sort_on_columns_list[1]
                c = sort_on_columns_list[2]
                sorted_data_text = sorted(data_text, key=lambda column: (int(column[a]) if isinstance(a, int) else column[int(a)], int(column[b]) if isinstance(b, int) else column[int(b)], int(column[c]) if isinstance(c, int) else column[int(c)]))
            case 4:
                a = sort_on_columns_list[0]
                b = sort_on_columns_list[1]
                c = sort_on_columns_list[2]
                d = sort_on_columns_list[3]
                sorted_data_text = sorted(data_text, key=lambda column: (int(column[a]) if isinstance(a, int) else column[int(a)], int(column[b]) if isinstance(b, int) else column[int(b)], int(column[c]) if isinstance(c, int) else column[int(c)], int(column[d]) if isinstance(d, int) else column[int(d)]))
            case 5:
                a = sort_on_columns_list[0]
                b = sort_on_columns_list[1]
                c = sort_on_columns_list[2]
                d = sort_on_columns_list[3]
                e = sort_on_columns_list[4]
                sorted_data_text = sorted(data_text, key=lambda column: (int(column[a]) if isinstance(a, int) else column[int(a)], int(column[b]) if isinstance(b, int) else column[int(b)], int(column[c]) if isinstance(c, int) else column[int(c)], int(column[d]) if isinstance(d, int) else column[int(d)], int(column[e]) if isinstance(e, int) else column[int(e)]))
            case 6:
                a = sort_on_columns_list[0]
                b = sort_on_columns_list[1]
                c = sort_on_columns_list[2]
                d = sort_on_columns_list[3]
                e = sort_on_columns_list[4]
                f = sort_on_columns_list[5]
                sorted_data_text = sorted(data_text, key=lambda column: (int(column[a]) if isinstance(a, int) else column[int(a)], int(column[b]) if isinstance(b, int) else column[int(b)], int(column[c]) if isinstance(c, int) else column[int(c)], int(column[d]) if isinstance(d, int) else column[int(d)], int(column[e]) if isinstance(e, int) else column[int(e)], int(column[f]) if isinstance(f, int) else column[int(f)]))
            case 7:
                a = sort_on_columns_list[0]
                b = sort_on_columns_list[1]
                c = sort_on_columns_list[2]
                d = sort_on_columns_list[3]
                e = sort_on_columns_list[4]
                f = sort_on_columns_list[5]
                g = sort_on_columns_list[6]
                sorted_data_text = sorted(data_text, key=lambda column: (int(column[a]) if isinstance(a, int) else column[int(a)], int(column[b]) if isinstance(b, int) else column[int(b)], int(column[c]) if isinstance(c, int) else column[int(c)], int(column[d]) if isinstance(d, int) else column[int(d)], int(column[e]) if isinstance(e, int) else column[int(e)], int(column[f]) if isinstance(f, int) else column[int(f)], int(column[g]) if isinstance(g, int) else column[int(g)]))
            case _:
                raise Exception("Too many columns selected for sorting. Only sorting on maximum 7 columns is supported.")
        return sorted_data_text
